"""

AutoProfile Script Demo
=======================

The following demo is end-to-end bash code that writes a demo script and
profiles it with autoprofile.

.. code:: bash

    # Write demo python script to disk
    python -c "if 1:
        import textwrap
        text = textwrap.dedent(
            '''
            def plus(a, b):
                return a + b

            def fib(n):
                a, b = 0, 1
                while a < n:
                    a, b = b, plus(a, b)

            def main():
                import math
                import time
                start = time.time()

                print('start calculating')
                while time.time() - start < 1:
                    fib(10)
                    math.factorial(1000)
                print('done calculating')

            main()
            '''
        ).strip()
        with open('demo.py', 'w') as file:
            file.write(text)
    "

    echo "---"
    echo "## Profile With AutoProfile"
    python -m kernprof -p demo.py -l demo.py
    python -m line_profiler -rmt demo.py.lprof
"""

import types
from .ast_tree_profiler import AstTreeProfiler
from .line_profiler_utils import add_imported_function_or_module

PROFILER_LOCALS_NAME = 'prof'


def _extend_line_profiler_for_profiling_imports(prof):
    """Allow profiler to handle functions/methods, classes & modules with a single call.

    Add a method to LineProfiler that can identify whether the object is a
    function/method, class or module and handle it's profiling accordingly.
    Mainly used for profiling objects that are imported.
    (Workaround to keep changes needed by autoprofile separate from base LineProfiler)

    Args:
        prof (LineProfiler):
            instance of LineProfiler.
    """
    prof.add_imported_function_or_module = types.MethodType(add_imported_function_or_module, prof)


def run(script_file, ns, prof_mod, profile_imports=False):
    """Automatically profile a script and run it.

    Profile functions, classes & modules specified in prof_mod without needing to add
    @profile decorators.

    Args:
        script_file (str):
            path to script being profiled.

        ns (dict):
            "locals" from kernprof scope.

        prof_mod (List[str]):
            list of imports to profile in script.
            passing the path to script will profile the whole script.
            the objects can be specified using its dotted path or full path (if applicable).

        profile_imports (bool):
            if True, when auto-profiling whole script, profile all imports aswell.
    """
    tree_profiled = AstTreeProfiler(script_file, prof_mod, profile_imports).profile()

    _extend_line_profiler_for_profiling_imports(ns[PROFILER_LOCALS_NAME])
    exec(compile(tree_profiled, script_file, 'exec'), ns, ns)



from typing import Dict
from typing import List
from typing import Callable
from _typeshed import Incomplete


class GlobalProfiler:
    setup_config: Dict[str, List[str]]
    output_prefix: str
    write_config: Dict[str, bool]
    show_config: Dict[str, bool]
    enabled: bool | None

    def __init__(self) -> None:
        ...

    def enable(self, output_prefix: Incomplete | None = ...) -> None:
        ...

    def disable(self) -> None:
        ...

    def __call__(self, func: Callable) -> Callable:
        ...

    def show(self) -> None:
        ...


profile: Incomplete



"""
Line Profiler
=============

The line_profiler module for doing line-by-line profiling of functions

+---------------+--------------------------------------------+
| Github        | https://github.com/pyutils/line_profiler   |
+---------------+--------------------------------------------+
| Pypi          | https://pypi.org/project/line_profiler     |
+---------------+--------------------------------------------+
| ReadTheDocs   | https://kernprof.readthedocs.io/en/latest/ |
+---------------+--------------------------------------------+


Installation
============

Releases of :py:mod:`line_profiler` and :py:mod:`kernprof` can be installed
using pip

.. code:: bash

    pip install line_profiler


The package also provides extras for optional dependencies, which can be
installed via:


.. code:: bash

    pip install line_profiler[all]


Line Profiler Basic Usage
=========================

To demonstrate line profiling, we first need to generate a Python script to
profile. Write the following code to a file called ``demo_primes.py``.

.. code:: python

    from line_profiler import profile


    @profile
    def is_prime(n):
        '''
        Check if the number "n" is prime, with n > 1.

        Returns a boolean, True if n is prime.
        '''
        max_val = n ** 0.5
        stop = int(max_val + 1)
        for i in range(2, stop):
            if n % i == 0:
                return False
        return True


    @profile
    def find_primes(size):
        primes = []
        for n in range(size):
            flag = is_prime(n)
            if flag:
                primes.append(n)
        return primes


    @profile
    def main():
        print('start calculating')
        primes = find_primes(100000)
        print(f'done calculating. Found {len(primes)} primes.')


    if __name__ == '__main__':
        main()


In this script we explicitly import the ``profile`` function from
``line_profiler``, and then we decorate function of interest with ``@profile``.

By default nothing is profiled when running the script.

.. code:: bash

    python demo_primes.py


The output will be

.. code::

    start calculating
    done calculating. Found 9594 primes.


The quickest way to enable profiling is to set the environment variable
``LINE_PROFILE=1`` and running your script as normal.


.. code:: bash

    LINE_PROFILE=1 python demo_primes.py

This will output 3 files: profile_output.txt, profile_output_<timestamp>.txt,
and profile_output.lprof and stdout will look something like:


.. code::

    start calculating
    done calculating. Found 9594 primes.
    Timer unit: 1e-09 s

      0.65 seconds - demo_primes.py:4 - is_prime
      1.47 seconds - demo_primes.py:19 - find_primes
      1.51 seconds - demo_primes.py:29 - main
    Wrote profile results to profile_output.txt
    Wrote profile results to profile_output_2023-08-12T193302.txt
    Wrote profile results to profile_output.lprof
    To view details run:
    python -m line_profiler -rtmz profile_output.lprof


For more control over the outputs, run your script using :py:mod:`kernprof`.
The following invocation will run your script, dump results to
``demo_primes.py.lprof``, and display results.

.. code:: bash

    python -m kernprof -lvr demo_primes.py


Note: the ``-r`` flag will use "rich-output" if you have the :py:mod:`rich`
module installed.

See Also:

    * autoprofiling usage in: :py:mod:`line_profiler.autoprofile`
"""
# Note: there are better ways to generate primes
# https://github.com/Sylhare/nprime

__submodules__ = [
    'line_profiler',
    'ipython_extension',
]

__autogen__ = """
mkinit ./line_profiler/__init__.py --relative
mkinit ./line_profiler/__init__.py --relative -w
"""


# from .line_profiler import __version__

# NOTE: This needs to be in sync with ../kernprof.py and line_profiler.py
__version__ = '4.1.1'

from .line_profiler import (LineProfiler,
                            load_ipython_extension, load_stats, main,
                            show_func, show_text,)


from .explicit_profiler import profile


__all__ = ['LineProfiler', 'line_profiler',
           'load_ipython_extension', 'load_stats', 'main', 'show_func',
           'show_text', '__version__', 'profile']



"""
Auto-Profiling
==============

In cases where you want to decorate every function in a script, module, or
package decorating every function with ``@profile`` can be tedious. To make
this easier "auto-profiling" was introduced to ``line_profiler`` in Version
4.1.0.

The "auto-profile" feature allows users to specify a script or module.
This is done by passing the name of script or module to kernprof.

To demonstrate auto-profiling, we first need to generate a Python script to
profile. Write the following code to a file called ``demo_primes2.py``.

.. code:: python

    def is_prime(n):
        '''
        Check if the number "n" is prime, with n > 1.

        Returns a boolean, True if n is prime.
        '''
        max_val = n ** 0.5
        stop = int(max_val + 1)
        for i in range(2, stop):
            if n % i == 0:
                return False
        return True


    def find_primes(size):
        primes = []
        for n in range(size):
            flag = is_prime(n)
            if flag:
                primes.append(n)
        return primes


    def main():
        print('start calculating')
        primes = find_primes(100000)
        print(f'done calculating. Found {len(primes)} primes.')


    if __name__ == '__main__':
        main()

Note that this is the nearly the same "primes" example from the "Basic Usage"
section in :py:mod:`line_profiler`, but none of the functions are decorated
with ``@profile``.

To run this script with auto-profiling we invoke kernprof with ``-p`` or
``--prof-mod`` and pass the names of the script. When running with ``-p`` we
must also run with ``-l`` to enable line profiling. In this example we include
``-v`` as well to display the output after we run it.

.. code:: bash

    python -m kernprof -lv -p demo_primes2.py demo_primes2.py


The output will look like this:

.. code::


    start calculating
    done calculating. Found 9594 primes.
    Wrote profile results to demo_primes2.py.lprof
    Timer unit: 1e-06 s

    Total time: 0.677348 s
    File: demo_primes2.py
    Function: is_prime at line 4

    Line #      Hits         Time  Per Hit   % Time  Line Contents
    ==============================================================
         4                                           def is_prime(n):
         5                                               '''
         6                                               Check if the number "n" is prime, with n > 1.
         7
         8                                               Returns a boolean, True if n is prime.
         9                                               '''
        10    100000      19921.6      0.2      2.9      math.floor(1.3)
        11    100000      17743.6      0.2      2.6      max_val = n ** 0.5
        12    100000      23962.7      0.2      3.5      stop = int(max_val + 1)
        13   2745693     262005.7      0.1     38.7      for i in range(2, stop):
        14   2655287     342216.1      0.1     50.5          if n % i == 0:
        15     90406      10401.4      0.1      1.5              return False
        16      9594       1097.2      0.1      0.2      return True


    Total time: 1.56657 s
    File: demo_primes2.py
    Function: find_primes at line 19

    Line #      Hits         Time  Per Hit   % Time  Line Contents
    ==============================================================
        19                                           def find_primes(size):
        20         1          0.3      0.3      0.0      primes = []
        21    100000      11689.5      0.1      0.7      for n in range(size):
        22    100000    1541848.0     15.4     98.4          flag = is_prime(n)
        23     90406      10260.0      0.1      0.7          if flag:
        24      9594       2775.9      0.3      0.2              primes.append(n)
        25         1          0.2      0.2      0.0      return primes


    Total time: 1.61013 s
    File: demo_primes2.py
    Function: main at line 28

    Line #      Hits         Time  Per Hit   % Time  Line Contents
    ==============================================================
        28                                           def main():
        29         1         17.6     17.6      0.0      print('start calculating')
        30         1    1610081.3    2e+06    100.0      primes = find_primes(100000)
        31         1         26.6     26.6      0.0      print(f'done calculating. Found {len(primes)} primes.')

"""
